/**
* TimrJS v3.0.0
* https://github.com/joesmith100/timrjs
* https://www.npmjs.com/package/timrjs
*
* Compatible with Browsers, Node.js (CommonJS) and RequireJS.
*
* Copyright (c) 2021 Joe Castle
* Released under the MIT license
* https://github.com/joesmith100/timrjs/blob/master/LICENSE.md
*/
(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.Timr = {}));
}(this, (function (exports) { 'use strict';

  /**
   * @description Pads beginning of number that are greater than or equal 0 and less than 10 with a 0.
   *
   * @param {number} num - the number to check
   *
   * @returns {string|number} The padded number or the original argument.
   */
  function zeroPad(num) {
      return (num >= 0 && num < 10) ? `0${num}` : num;
  }

  /**
   * @description Checks if a value is a valid number. Guards against NaN and Infinity.
   *
   * @param {any} value - the value to be checked
   *
   * @return {boolean} True if it is, false if it isn't
   */
  function isNum(value) {
      return typeof value === 'number' &&
          !isNaN(value) &&
          value !== Infinity &&
          value !== -Infinity;
  }
  /**
   * @description Checks if a value is a not a number.
   *
   * @param {any} value - the value to be checked
   *
   * @return {boolean} True if it is, false if it isn't
   */
  function isNotNum(value) {
      return !isNum(value);
  }
  /**
   * @description Checks if a value is a string.
   *
   * @param {any} value - the value to be checked
   *
   * @return {boolean} True if it is, false if it isn't
   */
  function isStr(value) {
      return typeof value === 'string';
  }
  /**
   * @description Checks if a value is not a string.
   *
   * @param {any} value - the value to be checked
   *
   * @return {boolean} True if it is, false if it isn't
   */
  function isNotStr(value) {
      return !isStr(value);
  }
  /**
   * @description Checks if a value is a boolean.
   *
   * @param {any} value - the value to be checked
   *
   * @return {boolean} True if it is, false if it isn't
   */
  function isBool(value) {
      return typeof value === 'boolean';
  }
  /**
   * @description Checks if a value is not a boolean.
   *
   * @param {any} value - the value to be checked
   *
   * @return {boolean} True if it is, false if it isn't
   */
  function isNotBool(value) {
      return !isBool(value);
  }
  /**
   * @description Checks if a value is a function.
   *
   * @param {any}  value - the value to be checked
   *
   * @return {boolean} True if it is, false if it isn't
   */
  function isFn(value) {
      return typeof value === 'function';
  }
  /**
   * @description Checks if a value is not a function.
   *
   * @param {any} value - the value to be checked
   *
   * @return {boolean} True if it is, false if it isn't
   */
  function isNotFn(value) {
      return !isFn(value);
  }
  /**
   * @description Checks if a value is an object. Guards against null and arrays.
   *
   * @param {any} value - the value to be checked
   *
   * @return {boolean} True if it is, false if it isn't
   */
  function isObj(value) {
      return value !== null &&
          typeof value === 'object' &&
          !Array.isArray(value);
  }
  /**
   * @description Checks if the provided object extends the provided class
   *
   * @param {value} value - the value to be checked
   * @param {any} clazz - the class to check against
   *
   * @returns True if it is, false if it isn't
   */
  function isInstanceOf(value, clazz) {
      return value instanceof clazz;
  }
  /**
   * @description Checks if a value exists, i.e. not undefined or null.
   *
   * @param {any} value - the value to be checked
   *
   * @return {boolean} True if it is, false if it isn't
   */
  function exists(value) {
      return value !== undefined && value !== null;
  }
  /**
   * @description Checks if a value doesn't exist, i.e is undefined.
   *
   * @param {any} value - the value to be checked
   *
   * @return {boolean} True if it is, false if it isn't
   */
  function notExists(value) {
      return !exists(value);
  }
  /**
   * @description A more useful type check than typeof.
   * Guards against gotchas like null and array being an object and NaN being a number.
   * For this purpose it also explicitly states Infinity as a type rather than a number.
   *
   * @param {any} value - The value to check the type of.
   *
   * @return {string|null} The type of the value.
   */
  function checkType(value) {
      if (value === undefined ||
          value === null ||
          value === Infinity ||
          value === -Infinity) {
          return `${value}`;
      }
      if (isNum(value))
          return 'number';
      if (isStr(value))
          return 'string';
      if (isBool(value))
          return 'boolean';
      if (isFn(value))
          return 'function';
      if (isObj(value))
          return 'object';
      if (Array.isArray(value))
          return 'array';
      if (isNaN(value))
          return `${NaN}`;
      return '';
  }

  const optionsFormatKeys = ['default', 'ss', 'SS', 'mm', 'MM', 'hh', 'HH', 'dd', 'DD'];
  const timeValues = optionsFormatKeys.slice(1);
  /**
   * @description Creates an object using the provided object of functions
   * and assigns it to all or their respective values.
   *
   * @param {Object} formatValues - The object of functions to assign to the time values.
   *
   * @return {Object} - The created object of functions.
   */
  function makeValues(formatValues) {
      return timeValues.reduce((obj, item) => ({
          ...obj,
          [item]: exists(formatValues[item]) ? formatValues[item] : exists(formatValues.default) ? formatValues.default : zeroPad
      }), {});
  }
  /**
   * @description Builds an options object from default and custom options.
   *
   * @param {Object} [newOptions] - Optional custom options.
   * @param {Object} [oldOptions] - Optional previous options.
   *
   * @throws If any option is invalid.
   *
   * @return {Object} Compiled options from default and custom.
   */
  function buildOptions(newOptions, oldOptions) {
      // Run through validation first, than create the options afterwards.
      if (exists(newOptions)) {
          const { formatOutput, countdown, formatValues } = newOptions;
          if (exists(formatOutput) && isNotStr(formatOutput)) {
              throw new TypeError(`Expected formatOutput to be a string; instead got: ${checkType(formatOutput)}`);
          }
          if (exists(countdown) && isNotBool(countdown)) {
              throw new TypeError(`Expected countdown to be a boolean; instead got: ${checkType(countdown)}`);
          }
          if (exists(formatValues)) {
              if (isObj(formatValues)) {
                  let toError = false;
                  let error = 'Expected formatValues to contain a list of keys with functions that return a string or number; instead got:\n';
                  /**
                   * Runs through each key to check that it:
                   *  - Is a valid property, see optionsFormatKeys above.
                   *  - It's value is a function and it returns either a string or number.
                   *
                   * It will then create a string for all errors it finds,
                   * than throw the error after its finished the loop, if any errors are found.
                   */
                  Object.keys(formatValues).forEach((key) => {
                      const value = formatValues[key];
                      if (!optionsFormatKeys.includes(key)) {
                          error += ` '${key}': is not a recognised property, should be one of: ${optionsFormatKeys.map(val => ` '${val}'`).toString().trim()}\n`;
                          toError = true;
                      }
                      else if (isFn(value)) {
                          if (isNotNum(value(5)) && isNotStr(value(5))) {
                              error += ` '${key}': the return type for this function is not a string or number, is: ${checkType(value(5))}\n`;
                              toError = true;
                          }
                      }
                      else {
                          error += ` '${key}': is not a function, is: ${checkType(value)}\n`;
                          toError = true;
                      }
                  });
                  if (toError) {
                      throw new Error(error);
                  }
              }
              else {
                  throw new TypeError(`Expected formatValues to be a an object of functions; instead got: ${checkType(formatValues)}`);
              }
          }
      }
      // Build formatValues option.
      if (exists(newOptions)) {
          const { formatValues } = newOptions;
          if (exists(formatValues)) {
              let newFormatValues;
              // If oldOptions provided, merge previous formatValues with new ones
              // Otherwise make new ones from newOptions.
              if (exists(oldOptions)) {
                  newFormatValues = makeValues({
                      ...oldOptions.formatValues,
                      ...formatValues
                  });
              }
              else {
                  newFormatValues = makeValues(formatValues);
              }
              newOptions.formatValues = newFormatValues;
          }
      }
      return {
          formatOutput: 'DD hh:{mm:ss}',
          countdown: true,
          formatValues: makeValues({}),
          ...oldOptions,
          ...newOptions
      };
  }

  /**
   * @description Converts seconds to time format.
   *
   * @param {Number} seconds - The seconds to convert.
   * @param {Object} [options] - The options to build the string.
   * @param {Boolean} [toBuild=true] - Specifies wether to run the buildOptions function or not.
   *
   * For example, if you don't have a fully formed options object to pass in, or none at all.
   * Primarily used in Timr.formatTime to avoid making unecassary calls to buildOptions on every tick.
   *
   * Setting to false will require `options.formatOuput` and `options.formatValues` to exist
   *
   * @return {Object} An object that that contains the formattedTime and the
   * raw values used to calculate the time.
   */
  function formatTime(seconds, options, toBuild = true) {
      /**
       * If options doesn't exist, build regardless.
       * If it does and toBuild = true, run buildOptions
       * Otherwise just return passed in options
       */
      const { formatOutput, formatValues } = (() => {
          if (isObj(options)) {
              if (toBuild || (isNotStr(options.formatOutput) && (isNotFn(options.formatValues)))) {
                  return buildOptions(options);
              }
              else {
                  return options;
              }
          }
          else {
              return buildOptions();
          }
      })();
      // @ts-expect-error
      const raw = {};
      raw.SS = seconds;
      raw.MM = Math.floor(raw.SS / 60);
      raw.HH = Math.floor(raw.MM / 60);
      raw.DD = Math.floor(raw.HH / 24);
      raw.ss = Math.floor(seconds % 60);
      raw.mm = Math.floor(raw.MM % 60);
      raw.hh = Math.floor(raw.HH % 24);
      raw.dd = raw.DD;
      let stringToFormat = formatOutput;
      const leftBracketPosition = stringToFormat.lastIndexOf('{') + 1;
      const rightBracketPosition = stringToFormat.indexOf('}');
      /**
       * Get values sitting inbetween { } (protectedValues)
       * Match returns empty strings for groups it can't find so filter them out.
       * Note: Regex will ALWAYS return a match of atleast one empty string, hence <string[]> cast
       */
      const protectedValues = stringToFormat
          .slice(leftBracketPosition, rightBracketPosition)
          .match(/DD|HH|MM|SS/gi);
      // If protectedValues exist, apply logic to removing them
      // otherwise, format string exactly as it appears
      if (protectedValues != null) {
          let lastValueAboveZero;
          if (raw.DD > 0)
              lastValueAboveZero = 'DD';
          else if (raw.HH > 0)
              lastValueAboveZero = 'HH';
          else if (raw.MM > 0)
              lastValueAboveZero = 'MM';
          else
              lastValueAboveZero = 'SS';
          /**
           * If the lastValueAboveZero is inbetween the protectedValues than don't remove them.
           * E.G. starting string: HH:{mm:ss} and lastValueAboveZero is 'ss' than begin slice
           * at leftBracketPosition rather than beginning at ss.
           */
          const tempSlice = stringToFormat.search(new RegExp(lastValueAboveZero, 'ig'));
          const beginSlice = tempSlice >= leftBracketPosition ? leftBracketPosition : tempSlice;
          stringToFormat = stringToFormat.slice(beginSlice >= 0 ? beginSlice : 0);
      }
      // Replaces all values in string with their respective number values
      const formattedTime = stringToFormat
          // Remove any remaining curly braces before formatting.
          .replace(/\{|\}/g, '')
          // Apply specific function form formatValues to value.
          .replace(/DD|HH|MM|SS/gi, (match) => formatValues[match](raw[match]));
      return { formattedTime, raw };
  }

  /**
   * @description Converts time format (HH:MM:SS) into seconds.
   *
   * Automatically rounds the returned number to avoid errors
   * with floating point values.
   *
   * If a pattern is provided (25h / 25m), than
   * it is converted before being validated and processed.
   *
   * @param {String|Number} time - The time to be converted.
   * If a number is provided it will simply return that number.
   *
   * @throws If the provided time is neither a number nor a string.
   * @throws If the provided time is a negative number.
   * @throws If the provided time is not in the correct format HH:MM:SS.
   *
   * @return {Number} - The time in seconds.
   */
  function timeToSeconds(time) {
      // If a positive number, skip processing and just return the rounded number.
      if (isNum(time) && time >= 0)
          return Math.round(time);
      let newTime = time;
      // Converts '25m', '25h' && '25d' into '25:00', '25:00:00' & '600:00:00' respectivley.
      if (isStr(newTime) && /^\d+[mhd]$/i.test(newTime)) {
          switch (newTime.slice(-1)) {
              case 'm':
                  newTime = newTime.replace(/^(\d+)m$/i, '$1:00');
                  break;
              case 'h':
                  newTime = newTime.replace(/^(\d+)h$/i, '$1:00:00');
                  break;
              case 'd':
                  newTime = newTime.replace(/^(\d+)d$/i, (m, g1) => `${parseInt(g1) * 24}:00:00`);
                  break;
          }
      }
      if (isNotStr(newTime) && isNotNum(newTime)) {
          throw new TypeError(`Expected time to be a string or number, instead got: ${checkType(newTime)}`);
      }
      // Checks if the string can't be converted to a number, i.e someone passed 10:00.
      // Then tests for the correct format
      if (isNaN(Number(newTime)) && isStr(newTime)) {
          if (!/^(\d+)(:\d+)?(:\d+)?$/.test(newTime)) {
              throw new SyntaxError(`Expected time to be in (hh:mm:ss) format, instead got: ${newTime}`);
          }
      }
      if (Number(newTime) < 0) {
          throw new Error(`Time cannot be a negative number, got: ${newTime}`);
      }
      return Math.round(newTime.split(':').reduce((prev, curr, index, arr) => {
          if (arr.length === 3) {
              if (index === 0)
                  return prev + (Number(curr) * 60 * 60);
              if (index === 1)
                  return prev + (Number(curr) * 60);
          }
          if (arr.length === 2) {
              if (index === 0)
                  return prev + (Number(curr) * 60);
          }
          return prev + Number(curr);
      }, 0));
  }

  /**
   * @description Confirms if provided startTime is a futureDate parseable by dateToSeconds
   *
   * @param startTime the startTime
   *
   * @returns True if it is, false otherwise
   */
  function isDateFormat(startTime) {
      return isInstanceOf(startTime, Date) || (isStr(startTime) && /^\d{4}-\d{2}-\d{2}/.test(startTime));
  }
  /**
   * @description Converts a date object or date string into seconds until that date/time.
   *
   * @param {string|number} startTime - The date object or date string.
   * @param {string|number} backupStartTime - startTime to use if provided startTime is in the past.
   *
   * @throws If the date matches the regex but is not the correct format.
   * @throws If the date is in the past.
   *
   * @return {Number} - Returns the converted seconds.
   */
  function dateToSeconds(startTime, backupStartTime) {
      // startTime in MS from epoch.
      let parsedStartTime;
      if (isInstanceOf(startTime, Date)) {
          parsedStartTime = startTime.getTime();
      }
      else if (isStr(startTime)) {
          const error = new Error('The provided date is not in the right format or is incorrect.\n' +
              'Expected a string in the format: YYYY-MM-DD[ HH:MM[:SS]].\n' +
              '(year)-(month)-(day) (hour):(minute):(second(s))\n' +
              'Time is optional and seconds is optional if time provided.\n' +
              `You passed: "${startTime}"`);
          try {
              // @ts-expect-error try catch will catch nullpointer failure
              const [year, month, ...rest] = startTime
                  .match(/^(\d{4})-(\d{2})-(\d{2})(?: (\d{2}):(\d{2})(?::(\d{2}))?)?$/i)
                  .slice(1)
                  .map((value) => parseInt(value))
                  .filter(Boolean);
              parsedStartTime = new Date(year, month - 1, ...rest).getTime();
              if (isNaN(parsedStartTime)) {
                  throw error;
              }
          }
          catch (err) {
              throw error;
          }
      }
      else {
          throw new Error(`Expected startTime to be a string or Date object, instead got: ${checkType(startTime)}`);
      }
      const dateNow = new Date();
      // startTime in seconds from now, using parsedStartTime above.
      const startTimeInSeconds = Math.ceil((parsedStartTime - dateNow.getTime()) / 1000);
      if (startTimeInSeconds < 0) {
          if (exists(backupStartTime)) {
              return dateToSeconds(backupStartTime);
          }
          throw new Error('When passing a date/time, it cannot be in the past. ' +
              `You passed: "${startTime}". It's currently: "` +
              `${zeroPad(dateNow.getFullYear())}-${zeroPad(dateNow.getMonth() + 1)}-` +
              `${zeroPad(dateNow.getDate())} ` +
              `${zeroPad(dateNow.getHours())}:${zeroPad(dateNow.getMinutes())}:` +
              `${zeroPad(dateNow.getSeconds())}"`);
      }
      return startTimeInSeconds;
  }

  /**
   * @description Creates an EventEmitter.
   *
   * This is a super slimmed down version of nodes EventEmitter.
   *
   * This is only intended for internal use, as there is
   * no real error checking.
   */
  class EventEmitter {
      constructor() {
          this.events = {};
      }
      /**
       * @description Registers a listener to an event array.
       *
       * @param {String} event - The event to attach to.
       * @param {Function} listener - The event listener.
       */
      on(event, listener) {
          if (notExists(this.events[event])) {
              this.events[event] = [];
          }
          this.events[event].push(listener);
      }
      /**
       * @description Emits an event, calling all listeners store
       * against the provided event.
       *
       * @param {String} event - The event to emit.
       * @param {Array} args - The functions to run against the event.
       */
      emit(event, ...args) {
          if (exists(this.events[event])) {
              this.events[event].forEach((listener) => {
                  listener.apply(this, args);
              });
          }
      }
      /**
       * @description Removes all listeners.
       */
      removeAllListeners() {
          this.events = {};
      }
  }

  /**
   * Timr
   */
  exports.Status = void 0;
  (function (Status) {
      Status["initialized"] = "initialized";
      Status["started"] = "started";
      Status["paused"] = "paused";
      Status["stopped"] = "stopped";
      Status["finished"] = "finished";
      Status["destroyed"] = "destroyed";
  })(exports.Status || (exports.Status = {}));

  class Timr extends EventEmitter {
      /**
       * @description Creates a Timr.
       *
       * @param {String|Number} startTime - The starting time for the timr object.
       * @param {Object} [options] - Options to customise the timer.
       *
       * @throws If the provided startTime is neither a number or a string,
       * or, incorrect format.
       */
      constructor(startTime, options) {
          super();
          // options needs to be built before startTime is set,
          // so it can work out the future date properly.
          this.changeOptions(options);
          this.setStartTime(startTime);
          this.status = exports.Status.initialized;
      }
      /**
       * @description Countdown function.
       *
       * Bound to a setInterval when start() is called.
       */
      _countdown() {
          this.currentTime -= 1;
          this.emit('ticker', {
              ...this.formatTime(),
              percentDone: this.percentDone(),
              currentTime: this.currentTime,
              startTime: this.startTime,
              self: this
          });
          if (this.currentTime <= 0) {
              this.clear();
              this.currentTime = this.startTime;
              this.emit('finish', this);
              this.status = exports.Status.finished;
          }
      }
      /**
       * @description Stopwatch function.
       *
       * Bound to a setInterval when start() is called.
       */
      _stopwatch() {
          this.currentTime += 1;
          this.emit('ticker', {
              ...this.formatTime(),
              currentTime: this.currentTime,
              startTime: this.startTime,
              self: this
          });
      }
      _listener(name, listener) {
          if (isNotFn(listener)) {
              throw new TypeError(`Expected ${name} to be a function, instead got: ${checkType(listener)}`);
          }
          this.on(name, listener);
          return this;
      }
      /**
       * @description Starts the timr.
       *
       * @param {Number} [delay] - Optional delay in ms to start the timer
       *
       * @return {Object} Returns a reference to the Timr so calls can be chained.
       */
      start(delay) {
          if (!this.started()) {
              if (this.options.countdown && this.startTime === 0) {
                  throw new Error('Unable to start timer when countdown = true and startTime = 0. ' +
                      'This would cause the timer to count into negative numbers and never stop. ' +
                      'Try setting countdown to false or amending the startTime.');
              }
              const startFn = () => {
                  /**
                   * futureDate records the original date used when futureDate option is set to true,
                   * this will re-run setStarTime to ensure the startTime is upto date.
                   *
                   * Note: Inside startFn so that delay works properly, if it was outside this scope,
                   * the startTime would be out of sync after the delay finishes.
                   */
                  if (exists(this.futureDate)) {
                      this.setStartTime(this.futureDate);
                  }
                  this.status = exports.Status.started;
                  this.timer = this.options.countdown
                      ? setInterval(() => this._countdown(), 1000)
                      : setInterval(() => this._stopwatch(), 1000);
              };
              if (exists(delay)) {
                  if (isNotNum(delay)) {
                      throw new TypeError(`The delay argument passed to start must be a number, you passed: ${checkType(delay)}`);
                  }
                  this.delayTimer = setTimeout(startFn, delay);
              }
              else {
                  startFn();
              }
              this.emit('onStart', this);
          }
          else {
              this.emit('onAlreadyStarted', this);
          }
          return this;
      }
      /**
       * @description Pauses the timr.
       *
       * @return {Object} Returns a reference to the Timr so calls can be chained.
       */
      pause() {
          this.clear();
          this.emit('onPause', this);
          this.status = exports.Status.paused;
          return this;
      }
      /**
       * @description Stops the timr.
       *
       * @return {Object} Returns a reference to the Timr so calls can be chained.
       */
      stop() {
          this.clear();
          this.currentTime = this.startTime;
          this.emit('onStop', this);
          this.status = exports.Status.stopped;
          return this;
      }
      /**
       * @description Clears the timr.
       *
       * @return {Object} Returns a reference to the Timr so calls can be chained.
       */
      clear() {
          clearInterval(this.timer);
          clearTimeout(this.delayTimer);
          return this;
      }
      /**
       * @description Destroys the timr,
       * clearing the interval, removing all event listeners and removing,
       * from the store (if it's in one).
       *
       * @return {Object} Returns a reference to the Timr so calls can be chained.
       */
      destroy() {
          this.emit('onDestroy', this);
          this.clear().removeAllListeners();
          // removeFromStore is added when the timr is added to a store,
          // so need to check if it's in a store before removing it.
          if (isFn(this.removeFromStore)) {
              this.removeFromStore();
          }
          this.status = exports.Status.destroyed;
          return this;
      }
      /**
       * @description Called every second the timer ticks down.
       *
       * @throws If the argument is not of type function.
       *
       * @param {Function} listener - Function to added to events.
       * @return {Object} Returns a reference to the Timr so calls can be chained.
       */
      ticker(listener) {
          return this._listener('ticker', listener);
      }
      /**
       * @description Called once when the timer finishes.
       *
       * @throws If the argument is not of type function.
       *
       * @param {Function} listener - Function to added to events.
       * @return {Object} Returns a reference to the Timr so calls can be chained.
       */
      finish(listener) {
          return this._listener('finish', listener);
      }
      /**
       * @description Called when the timer starts.
       *
       * @throws If the argument is not of type function.
       *
       * @param {Function} listener - Function to added to events.
       * @return {Object} Returns a reference to the Timr so calls can be chained.
       */
      onStart(listener) {
          return this._listener('onStart', listener);
      }
      /**
      * @description Called when the timer is already running and start is called
      *
      * @throws If the argument is not of type function.
      *
      * @param {Function} listener - Function to added to events.
      * @return {Object} Returns a reference to the Timr so calls can be chained.
      */
      onAlreadyStarted(listener) {
          return this._listener('onAlreadyStarted', listener);
      }
      /**
       * @description Called when the timer is paused.
       *
       * @throws If the argument is not of type function.
       *
       * @param {Function} listener - Function to added to events.
       * @return {Object} Returns a reference to the Timr so calls can be chained.
       */
      onPause(listener) {
          return this._listener('onPause', listener);
      }
      /**
       * @description Called when the timer is stopped.
       *
       * @throws If the argument is not of type function.
       *
       * @param {Function} listener - Function to added to events.
       * @return {Object} Returns a reference to the Timr so calls can be chained.
       */
      onStop(listener) {
          return this._listener('onStop', listener);
      }
      /**
       * @description Called when the timer is destroyed.
       *
       * @throws If the argument is not of type function.
       *
       * @param {Function} listener - Function to added to events.
       * @return {Object} Returns a reference to the Timr so calls can be chained.
       */
      onDestroy(listener) {
          return this._listener('onDestroy', listener);
      }
      /**
       * @description Converts seconds to time format.
       * This is provided to the ticker.
       *
       * @param {String} [time=currentTime] - option to format the startTime
       *
       * @return {Object} The formatted time and raw values.
       */
      formatTime(time = 'currentTime') {
          return formatTime(this[time], this.options, false);
      }
      /**
       * @description Returns the time elapsed in percent.
       * This is provided to the ticker.
       *
       * @return {Number} Time elapsed in percent.
       */
      percentDone() {
          return 100 - Math.round((this.currentTime / this.startTime) * 100);
      }
      /**
       * @description Creates / changes options for a Timr.
       * Merges with existing or default options.
       *
       * @param {Object} options - The options to create / change.
       * @return {Object} Returns a reference to the Timr so calls can be chained.
       */
      changeOptions(options) {
          this.options = buildOptions(options, this.options);
          return this;
      }
      /**
       * @description Sets new startTime after Timr has been created.
       * Will clear currentTime and reset to new startTime.
       *
       * @param {String|Number} startTime - The new start time.
       *
       * @throws If no startTime is provided.
       *
       * @return {Object} The original Timr object.
       */
      setStartTime(startTime, backupStartTime) {
          this.clear();
          if (notExists(startTime)) {
              throw new Error('You must provide a startTime value.');
          }
          let newStartTime;
          if (isDateFormat(startTime)) {
              newStartTime = dateToSeconds(startTime, backupStartTime ?? this.options.backupStartTime);
              this.futureDate = startTime;
          }
          else {
              newStartTime = timeToSeconds(startTime);
              this.futureDate = null;
          }
          this.startTime = newStartTime;
          this.currentTime = newStartTime;
          return this;
      }
      /**
       * @description Shorthand for this.formatTime(time).formattedTime
       */
      getFt(time = 'currentTime') {
          return this.formatTime(time).formattedTime;
      }
      /**
       * @description Shorthand for this.formatTime(time).raw
       */
      getRaw(time = 'currentTime') {
          return this.formatTime(time).raw;
      }
      /**
       * @description Gets the Timrs startTime.
       *
       * @return {Number} Start time in seconds.
       */
      getStartTime() {
          return this.startTime;
      }
      /**
       * @description Gets the Timrs currentTime.
       *
       * @return {Number} Current time in seconds.
       */
      getCurrentTime() {
          return this.currentTime;
      }
      /**
       * @description Returns the current status. Or returns
       * a boolean comparing the provided statusName to the current
       * status.
       *
       * @param {Status} [statusName] optional statusname to check
       *
       * @returns either the current status, or a boolean confirming the current status if
       * a statusName is provided. Will check `statusName === this.status`
       */
      getStatus(statusName) {
          return exists(statusName)
              ? this.status === statusName
              : this.status;
      }
      /**
       * @description Gets the Timrs running value.
       *
       * @deprecated please use `this.started()` instead
       *
       * @return {Boolean} True if running, false if not.
       */
      isRunning() {
          return this.started();
      }
      /**
       * @description Checks whether the timer has been started or not
       *
       * @return {Boolean} True if running, false if not.
       */
      started() {
          return this.getStatus(exports.Status.started);
      }
  }

  /**
   * @description Creates a store that can hold multiple timr objects
   * and perform functions on all of them.
   *
   * @param {Array} [args] - Optional timers to start the store with.
   * @throws If the provided timr is not a Timr object.
   * @throws If the provided timr is already in a store.
   *
   * @return {Object} Returns a store object with methods.
   */
  function createStore(...args) {
      let timrs = [];
      function removeFromStore(timr) {
          // Instanceof check required as it's exposed as a store method.
          if (timr instanceof Timr) {
              timrs = timrs.filter(x => x !== timr);
              timr.removeFromStore = null;
          }
      }
      /**
       * @description Adds the provided timr to the store.
       *
       * @param {Object} timr - A timr object.
       *
       * @throws If the provided timr is not a Timr object.
       * @throws If the provided timr is already in a store.
       *
       * @return {Object} The provided timr object.
       */
      function add(timr) {
          if (timr instanceof Timr && isNotFn(timr.removeFromStore)) {
              timrs.push(timr);
              timr.removeFromStore = () => {
                  removeFromStore(timr);
              };
          }
          else {
              throw new Error('Unable to add to store; provided argument is either already in a store ' +
                  'or not a timr object.');
          }
          return timr;
      }
      // Flatten args down to their values and add them to the store
      // if they pass validation.
      args.forEach(add);
      return {
          add,
          getAll: () => timrs,
          startAll: () => timrs.forEach(timr => timr.start()),
          pauseAll: () => timrs.forEach(timr => timr.pause()),
          stopAll: () => timrs.forEach(timr => timr.stop()),
          getStatus: (statusName) => timrs.filter(timr => timr.getStatus(statusName)),
          // Deprecated in favour of .started()
          isRunning: () => timrs.filter(timr => timr.isRunning()),
          started: () => timrs.filter(timr => timr.started()),
          removeFromStore,
          destroyAll: () => {
              timrs.forEach(timr => timr.destroy());
              timrs = [];
          }
      };
  }

  /**
     * @description Creates a new Timr object.
     *
     * @param {string|number|object} startTime - The starting time for the timr object.
     * Or an object with startTime and options.
     * @param {object} [options] - Options to customise the timer.
     *
     * @return {object} A new Timr object.
     */
  function create(startTime, options) {
      if (isObj(startTime)) {
          if (exists(startTime.startTime)) {
              return new Timr(startTime.startTime, startTime);
          }
          throw new Error('When providing only an object when creating a timer, it must have a startTime property.');
      }
      return new Timr(startTime, options);
  }

  exports.create = create;
  exports.createStore = createStore;
  exports.dateToSeconds = dateToSeconds;
  exports.formatTime = formatTime;
  exports.timeToSeconds = timeToSeconds;
  exports.zeroPad = zeroPad;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
