/// <reference types="node" />
import EventEmitter from './EventEmitter';
import { FormattedTime, Listener, OptionalOptions, Options, Raw, Status } from './types';
declare class Timr extends EventEmitter {
    timer: NodeJS.Timeout;
    delayTimer: NodeJS.Timeout;
    currentTime: number;
    startTime: number;
    options: Options;
    futureDate: string | number | Date | null;
    status: Status;
    removeFromStore?: null | (() => void);
    [key: string]: any;
    /**
     * @description Creates a Timr.
     *
     * @param {String|Number} startTime - The starting time for the timr object.
     * @param {Object} [options] - Options to customise the timer.
     *
     * @throws If the provided startTime is neither a number or a string,
     * or, incorrect format.
     */
    constructor(startTime: string | number | Date, options?: OptionalOptions);
    /**
     * @description Countdown function.
     *
     * Bound to a setInterval when start() is called.
     */
    _countdown(): void;
    /**
     * @description Stopwatch function.
     *
     * Bound to a setInterval when start() is called.
     */
    _stopwatch(): void;
    _listener(this: Timr, name: string, listener: Listener): Timr;
    /**
     * @description Starts the timr.
     *
     * @param {Number} [delay] - Optional delay in ms to start the timer
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    start(delay?: number): Timr;
    /**
     * @description Pauses the timr.
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    pause(): Timr;
    /**
     * @description Stops the timr.
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    stop(): Timr;
    /**
     * @description Clears the timr.
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    clear(): Timr;
    /**
     * @description Destroys the timr,
     * clearing the interval, removing all event listeners and removing,
     * from the store (if it's in one).
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    destroy(): Timr;
    /**
     * @description Called every second the timer ticks down.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener - Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    ticker(listener: Listener): Timr;
    /**
     * @description Called once when the timer finishes.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener - Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    finish(listener: Listener): Timr;
    /**
     * @description Called when the timer starts.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener - Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    onStart(listener: Listener): Timr;
    /**
    * @description Called when the timer is already running and start is called
    *
    * @throws If the argument is not of type function.
    *
    * @param {Function} listener - Function to added to events.
    * @return {Object} Returns a reference to the Timr so calls can be chained.
    */
    onAlreadyStarted(listener: Listener): Timr;
    /**
     * @description Called when the timer is paused.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener - Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    onPause(listener: Listener): Timr;
    /**
     * @description Called when the timer is stopped.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener - Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    onStop(listener: Listener): Timr;
    /**
     * @description Called when the timer is destroyed.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener - Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    onDestroy(listener: Listener): Timr;
    /**
     * @description Converts seconds to time format.
     * This is provided to the ticker.
     *
     * @param {String} [time=currentTime] - option to format the startTime
     *
     * @return {Object} The formatted time and raw values.
     */
    formatTime(time?: 'currentTime' | 'startTime'): FormattedTime;
    /**
     * @description Returns the time elapsed in percent.
     * This is provided to the ticker.
     *
     * @return {Number} Time elapsed in percent.
     */
    percentDone(): number;
    /**
     * @description Creates / changes options for a Timr.
     * Merges with existing or default options.
     *
     * @param {Object} options - The options to create / change.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    changeOptions(options?: OptionalOptions): Timr;
    /**
     * @description Sets new startTime after Timr has been created.
     * Will clear currentTime and reset to new startTime.
     *
     * @param {String|Number} startTime - The new start time.
     *
     * @throws If no startTime is provided.
     *
     * @return {Object} The original Timr object.
     */
    setStartTime(startTime: string | number | Date, backupStartTime?: string | Date): Timr;
    /**
     * @description Shorthand for this.formatTime(time).formattedTime
     */
    getFt(time?: 'currentTime' | 'startTime'): string;
    /**
     * @description Shorthand for this.formatTime(time).raw
     */
    getRaw(time?: 'currentTime' | 'startTime'): Raw;
    /**
     * @description Gets the Timrs startTime.
     *
     * @return {Number} Start time in seconds.
     */
    getStartTime(): number;
    /**
     * @description Gets the Timrs currentTime.
     *
     * @return {Number} Current time in seconds.
     */
    getCurrentTime(): number;
    /**
     * @description Returns the current status. Or returns
     * a boolean comparing the provided statusName to the current
     * status.
     *
     * @param {Status} [statusName] optional statusname to check
     *
     * @returns either the current status, or a boolean confirming the current status if
     * a statusName is provided. Will check `statusName === this.status`
     */
    getStatus(statusName?: Status): boolean | string;
    /**
     * @description Gets the Timrs running value.
     *
     * @deprecated please use `this.started()` instead
     *
     * @return {Boolean} True if running, false if not.
     */
    isRunning(): boolean;
    /**
     * @description Checks whether the timer has been started or not
     *
     * @return {Boolean} True if running, false if not.
     */
    started(): boolean;
}
export default Timr;
